function [phi,se] = mm_he(y,x,w,m,varargin)
% A method of moments estimator for the SARAR model in the heteroskedastic
% case.
if ~isempty(varargin)
    phi_t = varargin{1};
else
    phi_t = est_initial(y,x,w,m);
end
phi_t(1:2) = phi_t(1:2).*(abs(phi_t(1:2))<1) + 0.5*(abs(phi_t(1:2))>=1);

[n,kx] = size(x);
wy = w*y;
my = m*y;
mx = m*x;
mw = m*w;
mwy = m*wy;
info = struct('n',n,'w',w,'m',m,'wy',wy,'my',my,'mx',mx,'mw',mw,'mwy',mwy);
opts = optimoptions('fsolve','Algorithm','trust-region','SpecifyObjectiveGradient',true,'Display','off');
[phi1,~,exitflag] = fsolve(@obj_mm,phi_t(1:2),opts,y,x,info); % trust-region is better than quasi-newton.
if exitflag<0
    disp('No convergence in mm_he.m.');
end
rx = x - phi1(2)*mx; 
rx_sq = rx'*rx;
rsy = y-phi1(1)*wy-phi1(2)*my+phi1(1)*phi1(2)*mwy;
beta = rx_sq\(rx'*rsy);
phi = [phi1;beta];
if nargout>1
    mu = my-phi(1)*mwy-mx*beta;
    epsilon = y - phi(1)*wy - x*beta - phi(2)*mu;
    rwy = wy-phi(2)*mwy;
    de = -[rwy,mu,rx]; % The derivative of epsilon with respect to phi.
    r = speye(n) - phi(2)*m;
    mr = m/r;
    rs = r-phi1(1)*w+phi1(1)*phi1(2)*mw;
    wsr = w/rs;
    rwsr = r*wsr;
    rwswsr = ((w-phi1(2)*mw)/(speye(n) - phi1(1)*w))*wsr;
    dh1 = rwy'*de+epsilon'*[zeros(n,1),-mwy,zeros(n,kx)]-epsilon'*(diag(rwsr).*de)*2 ...
        -[epsilon'*(diag(rwswsr).*epsilon),epsilon'*(diag(-m*wsr+rwsr*mr).*epsilon),zeros(1,kx)];
    dh2 = mu'*de-epsilon'*[mwy,zeros(n,1),mx]-epsilon'*(diag(mr).*de)*2-[0,epsilon'*(diag(mr*mr).*epsilon),zeros(1,kx)];
    dh3 =  rx'*de-[zeros(kx,1),mx'*epsilon,zeros(kx)];
    dh = [dh1;dh2;dh3];
    g = rwsr - diag(diag(rwsr));
    t = mr - diag(diag(mr));
    rwsxb = r*(w*((speye(n)-phi(1)*w)\(x*beta)));
    e_sq = (epsilon.^2)'; % Row vector.
    g2Sigma = (g+g').*e_sq;
    t2Sigma = (t+t').*e_sq;
    Sigma_vec = e_sq'.*rwsxb;
    a11 = tr_ab(g2Sigma,g2Sigma)/2 + rwsxb'*Sigma_vec;
    a21 = tr_ab(t2Sigma,g2Sigma)/2;
    a22 = tr_ab(t2Sigma,t2Sigma)/2;
    a31 = rx'*Sigma_vec;
    a32 = zeros(kx,1);
    a33 = rx'*(e_sq'.*rx);
    Omega = [a11,a21',a31';a21,a22,a32';a31,a32,a33];
    se = sqrt(abs(diag((dh\Omega)/(dh'))));
end

%------------------------------------------------------------------
function fval = tr_ab(A,B)
% Compute the trace of AB.
fval = sum(A'.*B,'all');

%------------------------------------------------------------------
function [h,dh] = obj_mm(phi1,y,x,info)
% Objective function.
[n,w,m,wy,my,mx,mw,mwy] = deal(info.n,info.w,info.m,info.wy,info.my,info.mx,info.mw,info.mwy);
rx = x - phi1(2)*mx; % This rx1 depends on rho.
rx_sq = rx'*rx;
rsy = y-phi1(1)*wy-phi1(2)*my+phi1(1)*phi1(2)*mwy;
beta = rx_sq\(rx'*rsy);
epsilon = rsy - rx*beta;
mr = m/(speye(n) - phi1(2)*m);
rs = speye(n)-phi1(2)*m-phi1(1)*w+phi1(1)*phi1(2)*mw;
rwsr = (w-phi1(2)*mw)/rs;
mu = my-phi1(1)*mwy-mx*beta;
rwy = wy-phi1(2)*mwy;
h = [epsilon'*rwy-epsilon'*(diag(rwsr).*epsilon);epsilon'*mu-epsilon'*(diag(mr).*epsilon)];
if nargout>1 % Compute derivative.
    msy = my-phi1(1)*mwy;
    db_lam = -rx_sq\(rx'*rwy); % The derivative of beta with respect to lambda.
    db_rho = rx_sq\(mx'*(-rsy+rx*beta) + rx'*(-msy+mx*beta)); % The derivative of beta with respect to lambda.
    de = [-rwy-rx*db_lam,-msy+mx*beta-rx*db_rho]; % The derivative of epsilon with respect to phi.
    wsr = w/rs;
    rwswsr = ((w-phi1(2)*mw)/(speye(n) - phi1(1)*w))*wsr;
    dh1 = rwy'*de+epsilon'*[zeros(n,1),-mwy]-epsilon'*(diag(rwsr).*de)*2 ...
        -[epsilon'*(diag(rwswsr).*epsilon),epsilon'*(diag(-m*wsr+rwsr*mr).*epsilon)];
    dh2 = mu'*de-epsilon'*[mwy,zeros(n,1)]-epsilon'*mx*[db_lam,db_rho]-epsilon'*(diag(mr).*de)*2-[0,epsilon'*(diag(mr*mr).*epsilon)];
    dh = [dh1;dh2];
end