function [phi,se] = re_sarar(y,x,w,m,homo,varargin)
% Root estimator of the SARAR model.
% If varargin is supplied (any value), S^{-1} and R^{-1} are approximated, and
% varargin{1} is the order of approximation.

%% Compute some vectors and matrices needed for the root estimator.
[phi_t,wy,my,mwy,mx,m2,m3] = est_initial(y,x,w,m);
[n,kx] = size(x);
s = speye(n) - phi_t(1)*w;
r = speye(n) - phi_t(2)*m;
xb = x*phi_t(3:end);
u = y-phi_t(1)*wy-xb;
epsilon = u - phi_t(2)*my + phi_t(1)*phi_t(2)*mwy + mx*phi_t(3:end)*phi_t(2);
rwsxb = r*(w*(s\xb));
rx = x -phi_t(2)*mx;
mu = my-phi_t(1)*mwy-mx*phi_t(3:end);
rwy = wy - phi_t(2)*mwy;
ryxb = y - xb - phi_t(2)*my + mx*phi_t(3:end)*phi_t(2);
if isempty(varargin)
    [g,t] = invmat(phi_t,n,w,m,m2,m3,homo);
else
    [g,t] = invmat(phi_t,n,w,m,m2,m3,homo,varargin{1});
end

%% Compute the root estimator of rho.
de = -[rwy,rx]; % The derivative of epsilon with respect to [\lambda;\beta]'.
g2e = (g+g')*epsilon;
t2e = (t+t')*epsilon;
rxrwy = rx'*rwy;
rx2 = rx'*rx;
c = (t2e'*de)/[g2e'*de+rwsxb'*de;-[rxrwy,rx2]];
g2mu = (g+g')*mu;
t2mu = (t+t')*mu;
a1 = t2mu'*mu/2 - c*[mu'*g2mu/2;zeros(kx,1)];
b1 = -t2mu'*u + c*[g2mu'*u+mu'*rwsxb;rx'*mu];
c1 = u'*t*u - c*[u'*g*u+u'*rwsxb;rx'*u];
d1 = t2mu'*epsilon - c*[g2mu'*epsilon;zeros(kx,1)];
det = b1^2-4*a1*c1;
if det <= 0
    rho = -b1/(2*a1);
else
    rho = ((-b1-sqrt(det))/(2*a1))*(d1>0) + ((-b1+sqrt(det))/(2*a1))*(d1<=0);
end

%% Compute the root estimator of lambda.
de = -[mu,rx]; % The derivative of epsilon with respect to [\rho;\beta].
c = (g2e'*de+rwsxb'*de)/[t2e'*de;-[rx'*mu,rx2]];
g2rwy = (g+g')*rwy;
t2rwy = (t+t')*rwy;
rwyrwsxb = rwy'*rwsxb;
a2 = g2rwy'*rwy/2 - c*[rwy'*t2rwy/2;zeros(kx,1)];
b2 = -g2rwy'*ryxb - rwyrwsxb + c*[t2rwy'*ryxb;rxrwy];
c2 = ryxb'*(g*ryxb+rwsxb) - c*[ryxb'*t*ryxb;rx'*ryxb];
d2 = g2rwy'*epsilon + rwyrwsxb - c*[t2rwy'*epsilon;rxrwy];
det = b2^2-4*a2*c2;
if det <= 0
    lambda = -b2/(2*a2);
else
    lambda = ((-b2-sqrt(det))/(2*a2))*(d2>0) + ((-b2+sqrt(det))/(2*a2))*(d2<=0);
end

%% Compute beta.
rx = x - rho*mx;
beta = (rx'*rx)\(rx'*(y-lambda*wy - rho*my + rho*lambda*mwy));
phi = [lambda;rho;beta];

%% Compute standard error.
if nargout>1    
    epsilon = y - lambda*wy - x*beta - rho*my + lambda*rho*mwy + mx*beta*rho;
    de = -[wy-rho*mwy,my - lambda*mwy - mx*beta,rx]; % The derivative of epsilon with respect to [\lambda;\rho;\beta].
    if isempty(varargin)
        [g,t] = invmat(phi,n,w,m,m2,m3,homo); % Compute t, ws, and rwsr for the heteroskedastic case.
    else
        [g,t] = invmat(phi,n,w,m,m2,m3,homo,varargin{1});
    end
    rwsxb = (speye(n) - rho*m)*(w*((speye(n)-lambda*w)\(x*beta)));
    Gamma = [epsilon'*(g+g')*de+rwsxb'*de;epsilon'*(t+t')*de;rx'*de];
    if homo
        [sige,mu3,mu4] = deal(mean(epsilon.^2),mean(epsilon.^3),mean(epsilon.^4));
        dev = mu4-3*sige^2;
        a11 = sige^2*tr_ab(g+g',g)+ dev*diag(g)'*diag(g) + rwsxb'*rwsxb*sige + rwsxb'*diag(g)*mu3;
        a21 = sige^2*tr_ab(t+t',g) + dev*diag(t)'*diag(g) + rwsxb'*diag(t)*mu3;
        a22 = sige^2*tr_ab(t+t',t) + dev*diag(t)'*diag(t);
        a31 = rx'*(diag(g)*mu3 + rwsxb*sige);
        a32 = rx'*diag(t)*mu3;
        a33 = rx'*rx*sige;
    else
        e_sq = (epsilon.^2)'; % Row vector.
        g2Sigma = (g+g').*e_sq;
        t2Sigma = (t+t').*e_sq;
        Sigma_vec = e_sq'.*rwsxb;
        a11 = tr_ab(g2Sigma,g2Sigma)/2 + rwsxb'*Sigma_vec;
        a21 = tr_ab(g2Sigma,t2Sigma)/2;
        a22 = tr_ab(t2Sigma,t2Sigma)/2;
        a31 = rx'*Sigma_vec;
        a32 = zeros(kx,1);
        a33 = rx'*(e_sq'.*rx);
    end
    Omega = [a11,a21',a31';a21,a22,a32';a31,a32,a33];
    se = sqrt(abs(diag((Gamma\Omega)/(Gamma'))));
end

%------------------------------------------------------------------
function fval = tr_ab(A,B)
% Compute the trace of AB.
fval = sum(A'.*B,'all');

%------------------------------------------------------------------
function [g,t] = invmat(phi,n,w,m,m2,m3,homo,varargin)
% Compute G and T. varargin is used to indicate the order of
% approximation. No approximation is used if varargin is empty.
r = speye(n)-phi(2)*m;
mat_same = (sum(abs(w-m),'all')<1e-6); % Whether the two spatial weights matrices are the same.
if isempty(varargin)
    if mat_same
        mw = m2;
    else
        mw = m*w;
    end
    mr = m/r;
    rwsr = (w-phi(2)*mw)/(r-phi(1)*w+phi(1)*phi(2)*mw);
else
    order = varargin{1};
    [invS,invR,w_k,m_k] = deal(speye(n)); % invS: S^{-1}; invR: R^{-1}; w_k=W^k; m_k = M^k.
    k = 1;
    while k <= order
        if  k==1
            m_k = m;
        elseif k==2 
            m_k = m2; 
        elseif k==3
            m_k = m3;
        else
            m_k = m_k*m;
        end
        invR = invR + m_k*phi(2)^k;
        if mat_same; w_k = m_k; else; w_k = w_k*w; end
        invS = invS + w_k*phi(1)^k;
        k = k+1;
    end
    mr = m*invR;
    ws = w*invS;    
    rwsr = r*ws*invR;
end
if homo
    g = rwsr - trace(rwsr)*speye(n)/n;
    t = mr - trace(mr)*speye(n)/n;
else
    g = rwsr - diag(diag(rwsr));
    t = mr - diag(diag(mr));
end